@@ -1,26 +1,14 @@
 /* run-parts: run a bunch of scripts in a directory
  *
- * Debian Linux run-parts program
+ * Debian run-parts program
  * Copyright (C) 1996 Jeff Noxon <jeff@router.patch.net>,
- * Copyright (C) 1996-1998 Guy Maor <maor@debian.org>
+ * Copyright (C) 1996-1999 Guy Maor <maor@debian.org>
  *
  * This is free software; see the GNU General Public License version 2
  * or later for copying conditions.  There is NO warranty.
  *
  * Based on run-parts.pl version 0.2, Copyright (C) 1994 Ian Jackson.
  *
- * Revision History
- *
- * 01/07/96 - JEN - v0.3 - First version; released with miscutils-1.3-7
- *
- * 04/11/96 - GM  - v0.4 - Made it skip over directories (fixes Bug#2244)
- *                         Better error checking
- *
- * 06/01/96 - AND - v0.5 - Added --verbose option to print the name of each
- *                         script before executing it
- *
- * 01/20/98 - GM  - v0.6 - Added --report option.
- *
  */
 
 #include <stdio.h>
@@ -35,8 +23,11 @@
 #include <getopt.h>
 #include <string.h>
 #include <errno.h>
+#include <ctype.h>
+#include <signal.h>
+#include <sys/time.h>
 
-#define VERSION "1.10"
+#define VERSION "1.12"
 
 int test_mode = 0;
 int verbose_mode = 0;
@@ -61,7 +52,7 @@ void version ()
 {
   fprintf (stderr, "Debian GNU/Linux run-parts program, version " VERSION
 "\nCopyright (C) 1994 Ian Jackson, Copyright (C) 1996 Jeff Noxon.
-Copyright (C) 1996,1997,1998 Guy Maor
+Copyright (C) 1996,1997,1998,1999 Guy Maor
 This is free software; see the GNU General Public License version 2
 or later for copying conditions.  There is NO warranty.
 ");
@@ -103,18 +94,14 @@ void set_umask ()
 /* True or false? Is this a valid filename (upper/lower alpha, digits,
  * underscores, and hyphens only?)
  */
-int valid_name (char *filename)
+int valid_name (const struct dirent *d)
 {
-  while (*filename) {
-    if (!(((*filename >= 'a') && (*filename <= 'z')) ||
-	  ((*filename >= 'A') && (*filename <= 'Z')) ||
-	  ((*filename >= '0') && (*filename <= '9')) ||
-	  (*filename == '_') ||
-	  (*filename == '-')))
+  char *c = d->d_name;
+  while (*c) {
+    if (!isalnum(*c) && *c!='_' && *c!='-')
       return 0;
-    ++filename;
+    ++c;
   }
-  
   return 1;
 }
 
@@ -134,6 +121,13 @@ void set_fl (int fd, int flags)
 }
 
 
+/* We have to abort the select() call in run_part when we receive a
+ * SIGCHLD signal. We can't simply ignore it, so we do nothing.
+ */
+void catch_sigchld(int sig) {
+}
+
+
 /* Execute a file */
 void run_part (char *progname)
 {
@@ -168,9 +162,11 @@ void run_part (char *progname)
 
   if (report_mode) {
     fd_set set;
-    int max, r, printflag;
+    int max, r, w, printflag, errno_select;
     ssize_t c;
     char buf[4096];
+    struct timeval zerotime, *timeout;
+    struct sigaction sa, sold;
     
     close(pout[1]);
     close(perr[1]);
@@ -178,21 +174,23 @@ void run_part (char *progname)
     set_fl(perr[0], O_NONBLOCK);
     max = pout[0] > perr[0] ? pout[0]+1 : perr[0]+1;
     printflag = 0;
+    zerotime.tv_usec = 0;
+    zerotime.tv_sec = 0;
+    timeout = NULL;
+    sigaction(SIGCHLD, NULL, &sa);
+    sa.sa_handler = catch_sigchld;
+    
     for (;;) {
-      if (pout[0] < 0 && perr[0] < 0) {
-	r = waitpid(pid, &result, 0);
-	if (r > 0)
-	  break;
-	else {
-	  error("waitpid: %s", strerror(errno));
-	  exit(1);
-	}
-      }
-
       FD_ZERO(&set);
       if (pout[0] >= 0) FD_SET(pout[0], &set);
       if (perr[0] >= 0) FD_SET(perr[0], &set);
-      r = select(max, &set, 0, 0, 0);
+
+      /* select() call may be interrupted by SIGCHLD */
+      sigaction(SIGCHLD, &sa, &sold);
+      r = select(max, &set, 0, 0, timeout);
+      errno_select = errno;
+      sigaction(SIGCHLD, &sold, NULL);
+
       if (r > 0) {
 	if (pout[0] >= 0 && FD_ISSET(pout[0], &set)) {
 	  c = read(pout[0], buf, sizeof(buf));
@@ -226,13 +224,30 @@ void run_part (char *progname)
 	}
       }
       else if (r < 0) {
-	error("select: %s", strerror(errno));
-	exit(1);
+	/* select() was interrupted so check if we got a SIGCHLD
+	 * probably from our child (check this later).
+	 */
+	if (errno_select != EINTR) {
+	  error("select: %s", strerror(errno_select));
+	  exit(1);
+	}
       }
-      else {			/* should never happen */
+      else if (timeout == NULL) {	/* should never happen */
 	error("internal error: select() returned 0 !");
 	exit(1);
       }
+      if (timeout == NULL) {
+        w = waitpid(pid, &result, WNOHANG);
+        if (w > 0) {
+          /* Our child has died so prepare to flush the pipes */
+          timeout = &zerotime;
+	}  
+      }
+      else {
+	if (pout[0]) close(pout[0]);
+	if (perr[0]) close(perr[0]);
+	break;
+      }
     }
   }
   else {
@@ -255,46 +270,68 @@ void run_part (char *progname)
 void run_parts (char *dirname)
 {
   struct dirent **namelist;
-  char filename[PATH_MAX];
+  char *filename;
+  size_t filename_length, dirname_length;
   int entries, i, result;
   struct stat st;
 
+  /* dirname + "/" */
+  dirname_length = strlen(dirname) + 1;
+  /* dirname + "/" + ".." + "\0" (This will save one realloc.) */
+  filename_length = dirname_length + 2 + 1;
+  if (! (filename = malloc(filename_length))) {
+    error ("failed to allocate memory for path: %s", strerror(errno));
+    exit (1);
+  }
+  strcpy (filename, dirname);
+  strcat (filename, "/");
+
   /* scandir() isn't POSIX, but it makes things easy. */
-  entries = scandir (dirname, &namelist, 0, alphasort);
-  if (entries < 0)
+  entries = scandir (dirname, &namelist, valid_name, alphasort);
+  if (entries < 0) {
     error ("failed to open directory %s: %s", dirname, strerror (errno));
-
+    exit (1);
+  }
+  
   for (i = 0; i < entries; i++) {
-    if (valid_name (namelist[i]->d_name)) {
-      strcpy (filename, dirname);
-      strcat (filename, "/");
-      strcat (filename, namelist[i]->d_name);
-
-      result = stat (filename, &st);
-      if (result < 0) {
-	error ("failed to stat component %s: %s", filename,
-	       strerror (errno));
+    if (filename_length < dirname_length + strlen(namelist[i]->d_name) + 1) {
+      filename_length = dirname_length + strlen(namelist[i]->d_name) + 1;
+      if (!(filename = realloc(filename, filename_length))) {
+	error ("failed to reallocate memory for path: %s", strerror(errno));
 	exit (1);
       }
-      if (S_ISREG(st.st_mode) && !access (filename, X_OK)) {
-	if (test_mode)
-	  printf ("run-parts would run %s\n", filename);
-	else {
-	  if (verbose_mode)
-	    fprintf(stderr, "run-parts: executing %s\n", filename);
-	  run_part (filename);
-	}
+    }
+    strcpy (filename + dirname_length, namelist[i]->d_name);
+
+    strcpy (filename, dirname);
+    strcat (filename, "/");
+    strcat (filename, namelist[i]->d_name);
+    
+    result = stat (filename, &st);
+    if (result < 0) {
+      error ("failed to stat component %s: %s", filename,
+	     strerror (errno));
+      exit (1);
+    }
+    if (S_ISREG(st.st_mode) && !access (filename, X_OK)) {
+      if (test_mode)
+	printf ("run-parts would run %s\n", filename);
+      else {
+	if (verbose_mode)
+	  fprintf(stderr, "run-parts: executing %s\n", filename);
+	run_part (filename);
       }
-      else if (!S_ISDIR(st.st_mode)) {
-	printf ("run-parts: component %s is not an executable plain file\n",
+    }
+    else if (!S_ISDIR(st.st_mode)) {
+      printf ("run-parts: component %s is not an executable plain file\n",
 		filename);
 	exitstatus = 1;
-      }
-
     }
+
     free (namelist[i]);
   }
   free (namelist);
+  free (filename);
 }
 
 /* Process options */
